==> moe-0.3.tar.bz2: moe-0.3/window.cc <==
/*  Moe - My Own Editor
    Copyright (C) 2005 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <algorithm>
#include <string>
#include <vector>

#include "buffer.h"
#include "block.h"
#include "iso_8859.h"
#include "menu.h"
#include "rc.h"
#include "regex.h"
#include "screen.h"
#include "window.h"


Window::Window( const int tl, const int h, Buffer_handle & bh ) throw()
  : _top_line( tl ), _height( h ), bhp( &bh )
  {
  bhp->buffer().pvalid( bhp->pointer );
  update_points( bhp->pointer, false );
  }


Point Window::clock_position() const throw()
  { return Point( _top_line, Screen::width() - 20 ); }


Point Window::absolute_cursor() const throw()	// cursor position on screen
  { return relative_cursor() + Point( _top_line, 0 ); }


Point Window::relative_cursor() const throw()	// cursor position on window
  { return Point( 1, 0 ) + bhp->cursor - bhp->top_left; }


void Window::center_cursor() throw()
  {
  bhp->top_left.line = bhp->buffer().lines();
  update_points( bhp->pointer, true, true );
  }


void Window::goto_bof() throw()
  {
  Point bof = bhp->buffer().bof();
  if( bhp->pointer != bof ) update_points( bof );
  }


void Window::goto_eof() throw()
  {
  Point eof = bhp->buffer().eof();
  if( bhp->pointer != eof || bhp->cursor != bhp->pcursor )
    update_points( eof );
  }


void Window::goto_home() throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();
  if( bhp->cursor != bhp->pcursor || p != buffer.bol( p ) )
    p = buffer.bol( p );
  else
    {
    if( !RC::options().smart_home || p == buffer.bot( p ) ) return;
    p = buffer.bot( p );
    }
  update_points( p );
  }


void Window::goto_eol() throw()
  {
  Point eol = bhp->buffer().eol( bhp->pointer );
  if( bhp->cursor != bhp->pcursor || bhp->pointer != eol )
    update_points( eol );
  }


void Window::goto_begin_of_block() throw()
  {
  if( Block::same_buffer( buffer() ) && buffer().pisvalid( Block::begin() ) )
    update_points( Block::begin(), true, true );
  else Screen::show_message( "Begin of block not set in this file" );
  }


void Window::goto_end_of_block() throw()
  {
  if( Block::same_buffer( buffer() ) && buffer().pisvalid( Block::end() ) )
    update_points( Block::end(), true, true );
  else Screen::show_message( "End of block not set in this file" );
  }


void Window::goto_line() throw()
  {
  static std::vector< std::string > history;

  history.push_back( std::string() );
  if( Screen::get_string( "Go to line (^C to abort): ", history ) <= 0 )
    return;
  std::string & s = history.back();
  Point p( 0, 0 );
  if( !RC::parse_int( s, p.line ) || p.line <= 0 )
    { Screen::show_message( "Invalid line number" ); history.pop_back();
    return; }
  --p.line;
  if( !bhp->buffer().pisvalid( p ) ) p = bhp->buffer().eof();
  update_points( p, true, true );
  }


void Window::goto_column() throw()
  {
  static std::vector< std::string > history;

  history.push_back( std::string() );
  if( Screen::get_string( "Go to column (^C to abort): ", history ) <= 0 )
    return;
  std::string & s = history.back();
  Point c = bhp->cursor;
  if( !RC::parse_int( s, c.col ) || c.col <= 0 )
    { Screen::show_message( "Invalid column number" ); history.pop_back();
    return; }
  --c.col;
  update_points( c, true, false, from_cursor );
  }


void Window::goto_mark( const int i ) throw()
  {
  if( i >= 0 && i <= 9 )
    {
    const Point & p = buffer().mark[i];
    if( bhp->buffer().pisvalid( p ) )
      update_points( p, true, true );
    else
      {
      char buf[32];
      std::snprintf( buf, sizeof( buf ), "Mark %d not set", i );
      Screen::show_message( buf );
      }
    }
  }


void Window::goto_matching_delimiter( const bool forward ) throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();

  if( buffer.set_to_matching_delimiter( p, forward ) )
    { update_points( p ); return; }
  if( p.col > 0 )
    {
    if( p == buffer.eol( p ) )
      {
      --p.col;
      if( buffer.set_to_matching_delimiter( p, forward ) )
        { update_points( p ); return; }
      }
    char ch2 = buffer[p], ch1 = buffer.ppgetc( p );
    if( ( ch1 == '/' && ch2 == '*' ) || ( ch1 == '*' && ch2 == '/' ) )
      if( buffer.set_to_matching_delimiter( p, forward ) )
        { update_points( p ); return; }
    }
  Screen::show_message( "No matching delimiter" );
  }


void Window::goto_pnext() throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();
  if( buffer.pnext( p ) || bhp->cursor != bhp->pcursor )
    update_points( p );
  }


void Window::goto_pprev() throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();
  if( bhp->cursor != bhp->pcursor || buffer.pprev( p ) )
    update_points( p );
  }


void Window::move_page( const bool down ) throw()
  {
  const int keep_lines = RC::options().keep_lines;
  int lines = ( _height - 1 ) / 2;

  if( keep_lines >= 0 && keep_lines <= lines )
    lines = _height - 1 - keep_lines;
  if( !down ) lines = -lines;
  move_vertical( lines );
  }


void Window::move_vertical( const int lines ) throw()
  {
  Point c = bhp->cursor;

  if( lines < 0 )
    { if( c.line == 0 ) { return; } c.line += lines; c.cut(); }
  else if( lines > 0 )
    {
    int last_line = bhp->buffer().last_line();
    if( c.line == last_line ) return;
    c.line += lines; if( c.line > last_line ) c.line = last_line;
    }
  else return;

  update_points( c, true, false, from_cursor );
  }


void Window::scroll_horizontal( const int cols ) throw()
  {
  Point tl = bhp->top_left;

  if( cols < 0 ) { if( tl.col == 0 ) { return; } tl.col += cols; tl.cut(); }
  else if( cols > 0 ) tl.col += cols;
  else return;

  update_points( tl, true, false, from_top_left );
  }


void Window::scroll_page( const bool down ) throw()
  {
  const int keep_lines = RC::options().keep_lines;
  int lines = ( _height - 1 ) / 2;

  if( keep_lines >= 0 && keep_lines < lines )
    lines = _height - 1 - keep_lines;
  if( !down ) lines = -lines;
  scroll_vertical( lines );
  }


void Window::scroll_vertical( const int lines ) throw()
  {
  Point tl = bhp->top_left;

  if( lines < 0 ) { if( tl.line == 0 ) { return; } tl.line += lines; tl.cut(); }
  else if( lines > 0 )
    {
    int first_line = bhp->buffer().last_line() - ( _height - 2 );
    if( first_line < 0 || tl.line == first_line ) return;
    tl.line += lines; if( tl.line > first_line ) tl.line = first_line;
    }
  else return;

  update_points( tl, true, false, from_top_left );
  }


void Window::set_mark( const int i ) throw()
  {
  if( i >= 0 && i <= 9 )
    {
    buffer().mark[i] = bhp->pointer;
    char buf[32];
    std::snprintf( buf, sizeof( buf ), "Mark %d set", i );
    Screen::show_message( buf );
    }
  }


void Window::repaint() const throw()
  {
  show_status_line();
  for( int i = 0; i < _height - 1; ++i )
    {
    Screen::out_buffer_line( bhp->buffer(), bhp->top_left + Point( i, 0 ),
                             _top_line + i + 1 );
    }
  }


void Window::show_character_info() const throw()
  {
  Screen::show_message( bhp->buffer().character_info( bhp->pointer ) );
  }


// prefix == keboard status ( ^K, ^Q, ^[, etc )
//
void Window::show_status_line( const char * prefix ) const throw()
  {
  if( _top_line < 0 ) return;
  char buf[Screen::width()+3];
  int len = std::snprintf( buf, sizeof( buf ), "@i%-3s %c%c%c%c%c%c %s %s",
                           prefix,
                           bhp->buffer().options.overwrite ? 'O' : 'I',
                           bhp->buffer().options.word_wrap ? 'W' : ' ',
                           bhp->buffer().options.auto_indent ? 'A' : ' ',
                           bhp->buffer().options.read_only ? 'R' : ' ',
                           /*RC::options().rectangle_mode ? 'X' :*/ ' ',
                           bhp->buffer().modified() ? '*' : ' ',
                           bhp->buffer().uname().c_str(),
                           bhp->buffer().modified() ? "(Modified)" : "" );
  int offset = Screen::width() - 40 + 2;	// +2 for the '@i'
  const Point & pc = bhp->pcursor;
  for( int i = pc.line + 1; i > 9999; i /= 10 ) --offset;
  for( int i = pc.col + 1; i > 999; i /= 10 ) --offset;
  if( len < offset ) std::memset( buf + len, ' ', offset - len );
  std::snprintf( buf + offset, sizeof( buf ) - offset,
                 "  Line %-4d Col %-3d 00:00  F1 for help",
                 pc.line + 1, pc.col + 1 );
  Screen::out_line( buf, _top_line );
  }


void Window::update_points( const Point & to, const bool show,
                    bool center, const From from ) throw()
  {
  Buffer_handle & bh = buffer_handle();
  const Buffer & buffer = bh.buffer();
  Point & tl = bh.top_left;
  Point & c = bh.cursor;
  Point & p = bh.pointer;
  Point & pc = bh.pcursor;
  Point old_tl = tl;

  switch( from )
    {
    case from_top_left:
      { Point dif = c - tl;
      tl = to; tl.cut(); c = tl + dif; p = buffer.to_pointer( c, pc );
      center = false; break; }
    case from_cursor: c = to; c.cut(), p = buffer.to_pointer( c, pc ); break;
    case from_pointer: p = to; c = pc = buffer.to_cursor( p ); break;
    }

  Point rc = relative_cursor();
  if( rc.line >= _height - 1 )
    {
    int d = center ? ( _height - 1 ) / 2 : _height - 2;
    tl.line = c.line - d; tl.cut();
    }
  else if( rc.line < 1 )
    {
    tl.line = c.line;
    if( center ) { tl.line -= ( _height - 1 ) / 2; tl.cut(); }
    }
  if( rc.col >= Screen::width() ) tl.col = c.col - Screen::width() + 1;
  else if( rc.col < 1 ) tl.col = c.col;

  if( show ) { if( tl == old_tl ) show_status_line(); else repaint(); }
  }

==> moe-0.4.tar.bz2: moe-0.4/window.cc <==
/*  GNU Moe - My Own Editor
    Copyright (C) 2005 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <algorithm>
#include <string>
#include <vector>

#include "buffer.h"
#include "block.h"
#include "iso_8859.h"
#include "menu.h"
#include "rc.h"
#include "regex.h"
#include "screen.h"
#include "window.h"


Window::Window( const int tl, const int h, Buffer_handle & bh ) throw()
  : _top_line( tl ), _height( h ), bhp( &bh )
  {
  bhp->buffer().pvalid( bhp->pointer );
  update_points( bhp->pointer, false );
  }


Point Window::clock_position() const throw()
  { return Point( _top_line, Screen::width() - 20 ); }


Point Window::absolute_cursor() const throw()	// cursor position on screen
  { return relative_cursor() + Point( _top_line, 0 ); }


Point Window::relative_cursor() const throw()	// cursor position on window
  { return Point( 1, 0 ) + bhp->cursor - bhp->top_left; }


void Window::center_cursor() throw()
  {
  bhp->top_left.line = bhp->buffer().lines();
  update_points( bhp->pointer, true, true );
  }


void Window::goto_bof() throw()
  {
  Point bof = bhp->buffer().bof();
  if( bhp->pointer != bof ) update_points( bof );
  }


void Window::goto_eof() throw()
  {
  Point eof = bhp->buffer().eof();
  if( bhp->pointer != eof || bhp->cursor != bhp->pcursor )
    update_points( eof );
  }


void Window::goto_home() throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();
  if( bhp->cursor != bhp->pcursor || p != buffer.bol( p ) )
    p = buffer.bol( p );
  else
    {
    if( !RC::options().smart_home || p == buffer.bot( p ) ) return;
    p = buffer.bot( p );
    }
  update_points( p );
  }


void Window::goto_eol() throw()
  {
  Point eol = bhp->buffer().eol( bhp->pointer );
  if( bhp->cursor != bhp->pcursor || bhp->pointer != eol )
    update_points( eol );
  }


void Window::goto_begin_of_block() throw()
  {
  if( Block::same_buffer( buffer() ) && buffer().pisvalid( Block::begin() ) )
    update_points( Block::begin(), true, true );
  else Screen::show_message( "Begin of block is not in this file" );
  }


void Window::goto_end_of_block() throw()
  {
  if( Block::same_buffer( buffer() ) && buffer().pisvalid( Block::end() ) )
    update_points( Block::end(), true, true );
  else Screen::show_message( "End of block is not in this file" );
  }


void Window::goto_line() throw()
  {
  static std::vector< std::string > history;

  history.push_back( std::string() );
  if( Screen::get_string( "Go to line (^C to abort): ", history ) <= 0 )
    return;
  std::string & s = history.back();
  Point p( 0, 0 );
  if( !RC::parse_int( s, p.line ) || p.line <= 0 )
    { Screen::show_message( "Invalid line number" ); history.pop_back();
    return; }
  --p.line;
  if( !bhp->buffer().pisvalid( p ) ) p = bhp->buffer().eof();
  update_points( p, true, true );
  }


void Window::goto_column() throw()
  {
  static std::vector< std::string > history;

  history.push_back( std::string() );
  if( Screen::get_string( "Go to column (^C to abort): ", history ) <= 0 )
    return;
  std::string & s = history.back();
  Point c = bhp->cursor;
  if( !RC::parse_int( s, c.col ) || c.col <= 0 )
    { Screen::show_message( "Invalid column number" ); history.pop_back();
    return; }
  --c.col;
  update_points( c, true, false, from_cursor );
  }


void Window::goto_mark( const int i ) throw()
  {
  if( i >= 0 && i <= 9 )
    {
    const Point & p = buffer().mark[i];
    if( bhp->buffer().pisvalid( p ) )
      update_points( p, true, true );
    else
      {
      char buf[32];
      std::snprintf( buf, sizeof( buf ), "Mark %d not set", i );
      Screen::show_message( buf );
      }
    }
  }


void Window::goto_matching_delimiter( const bool forward ) throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();

  if( buffer.set_to_matching_delimiter( p, forward ) )
    { update_points( p ); return; }
  if( p.col > 0 )
    {
    if( p == buffer.eol( p ) )
      {
      --p.col;
      if( buffer.set_to_matching_delimiter( p, forward ) )
        { update_points( p ); return; }
      }
    char ch2 = buffer[p], ch1 = buffer.ppgetc( p );
    if( ( ch1 == '/' && ch2 == '*' ) || ( ch1 == '*' && ch2 == '/' ) )
      if( buffer.set_to_matching_delimiter( p, forward ) )
        { update_points( p ); return; }
    }
  Screen::show_message( "No matching delimiter" );
  }


void Window::goto_pnext() throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();
  if( buffer.pnext( p ) || bhp->cursor != bhp->pcursor )
    update_points( p );
  }


void Window::goto_pprev() throw()
  {
  Point p = bhp->pointer;
  const Buffer & buffer = bhp->buffer();
  if( bhp->cursor != bhp->pcursor || buffer.pprev( p ) )
    update_points( p );
  }


void Window::move_page( const bool down ) throw()
  {
  const int keep_lines = RC::options().keep_lines;
  int lines = ( _height - 1 ) / 2;

  if( keep_lines >= 0 && keep_lines <= lines )
    lines = _height - 1 - keep_lines;
  if( !down ) lines = -lines;
  move_vertical( lines );
  }


void Window::move_vertical( const int lines ) throw()
  {
  Point c = bhp->cursor;

  if( lines < 0 )
    { if( c.line == 0 ) { return; } c.line += lines; c.cut(); }
  else if( lines > 0 )
    {
    int last_line = bhp->buffer().last_line();
    if( c.line == last_line ) return;
    c.line += lines; if( c.line > last_line ) c.line = last_line;
    }
  else return;

  update_points( c, true, false, from_cursor );
  }


void Window::scroll_horizontal( const int cols ) throw()
  {
  Point tl = bhp->top_left;

  if( cols < 0 ) { if( tl.col == 0 ) { return; } tl.col += cols; tl.cut(); }
  else if( cols > 0 ) tl.col += cols;
  else return;

  update_points( tl, true, false, from_top_left );
  }


void Window::scroll_page( const bool down ) throw()
  {
  const int keep_lines = RC::options().keep_lines;
  int lines = ( _height - 1 ) / 2;

  if( keep_lines >= 0 && keep_lines < lines )
    lines = _height - 1 - keep_lines;
  if( !down ) lines = -lines;
  scroll_vertical( lines );
  }


void Window::scroll_vertical( const int lines ) throw()
  {
  Point tl = bhp->top_left;

  if( lines < 0 ) { if( tl.line == 0 ) { return; } tl.line += lines; tl.cut(); }
  else if( lines > 0 )
    {
    int first_line = bhp->buffer().last_line() - ( _height - 2 );
    if( first_line < 0 || tl.line == first_line ) return;
    tl.line += lines; if( tl.line > first_line ) tl.line = first_line;
    }
  else return;

  update_points( tl, true, false, from_top_left );
  }


void Window::set_mark( const int i ) throw()
  {
  if( i >= 0 && i <= 9 )
    {
    buffer().mark[i] = bhp->pointer;
    char buf[32];
    std::snprintf( buf, sizeof( buf ), "Mark %d set", i );
    Screen::show_message( buf );
    }
  }


void Window::repaint() const throw()
  {
  show_status_line();
  for( int i = 0; i < _height - 1; ++i )
    {
    Screen::out_buffer_line( bhp->buffer(), bhp->top_left + Point( i, 0 ),
                             _top_line + i + 1 );
    }
  }


void Window::show_character_info() const throw()
  {
  Screen::show_message( bhp->buffer().character_info( bhp->pointer ) );
  }


// prefix == keboard status ( ^K, ^Q, ^[, etc )
//
void Window::show_status_line( const char * prefix ) const throw()
  {
  if( _top_line < 0 ) return;
  char buf[Screen::width()+3];
  int len = std::snprintf( buf, sizeof( buf ), "@i%-3s %c%c%c%c%c%c %s %s",
                           prefix,
                           bhp->buffer().options.overwrite ? 'O' : 'I',
                           bhp->buffer().options.word_wrap ? 'W' : ' ',
                           bhp->buffer().options.auto_indent ? 'A' : ' ',
                           bhp->buffer().options.read_only ? 'R' : ' ',
                           /*RC::options().rectangle_mode ? 'X' :*/ ' ',
                           bhp->buffer().modified() ? '*' : ' ',
                           bhp->buffer().uname().c_str(),
                           bhp->buffer().modified() ? "(Modified)" : "" );
  int offset = Screen::width() - 40 + 2;	// +2 for the '@i'
  const Point & pc = bhp->pcursor;
  for( int i = pc.line + 1; i > 9999; i /= 10 ) --offset;
  for( int i = pc.col + 1; i > 999; i /= 10 ) --offset;
  if( len < offset ) std::memset( buf + len, ' ', offset - len );
  std::snprintf( buf + offset, sizeof( buf ) - offset,
                 "  Line %-4d Col %-3d 00:00  F1 for help",
                 pc.line + 1, pc.col + 1 );
  Screen::out_line( buf, _top_line );
  }


void Window::update_points( const Point & to, const bool show,
                    bool center, const From from ) throw()
  {
  Buffer_handle & bh = buffer_handle();
  const Buffer & buffer = bh.buffer();
  Point & tl = bh.top_left;
  Point & c = bh.cursor;
  Point & p = bh.pointer;
  Point & pc = bh.pcursor;
  Point old_tl = tl;

  switch( from )
    {
    case from_top_left:
      { Point dif = c - tl;
      tl = to; tl.cut(); c = tl + dif; p = buffer.to_pointer( c, pc );
      center = false; break; }
    case from_cursor: c = to; c.cut(), p = buffer.to_pointer( c, pc ); break;
    case from_pointer: p = to; c = pc = buffer.to_cursor( p ); break;
    }

  const Point rc = relative_cursor();
  if( rc.line >= _height - 1 )
    {
    const int d = center ? ( _height - 1 ) / 2 : _height - 2;
    tl.line = std::max( 0, c.line - d );
    }
  else if( rc.line < 1 )
    {
    tl.line = c.line;
    if( center ) { tl.line -= ( _height - 1 ) / 2; tl.cut(); }
    }
  if( rc.col >= Screen::width() ) tl.col = c.col - Screen::width() + 1;
  else if( rc.col < 0 ) tl.col = std::max( 0, c.col - ( Screen::width() / 2 ) );

  if( show ) { if( tl == old_tl ) show_status_line(); else repaint(); }
  }
